

/* 
 * Argyll Color Correction System
 * Common display patch reading support.
 *
 * Author: Graeme W. Gill
 * Date:   2/11/2005
 *
 * Copyright 1998 - 2005 Graeme W. Gill
 * All rights reserved.
 *
 * This material is licenced under the GNU GENERAL PUBLIC LICENCE :-
 * see the Licence.txt file for licencing details.
 */

#ifdef __MINGW32__
# define WINVER 0x0500
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include "copyright.h"
#include "config.h"
#include "numlib.h"
#include "xspect.h"
#include "insttypes.h"
#include "icoms.h"
#include "inst.h"
#include "dtp92.h"
#include "gretag.h"
#include "dispwin.h"
#include "dispsup.h"

#undef DEBUG

/* User requested calibration of the instrument */
int disprd_calibration(
instType itype,		/* Instrument type */
int comport, 		/* COM port used */
int dtype,			/* Display type, 1 = CRT, 2 = LCD */
int nocal,			/* NZ to disable auto instrument calibration */
disppath *disp,		/* display to calibrate. */
int override,		/* Override_redirect on X11 */
double patsize,		/* Size of dispwin */
double ho, double vo,	/* Position of dispwin */
int verb,			/* Verbosity flag */
int debug			/* Debug flag */
) {
	inst *it;
	int c;
	inst_code rv;
	baud_rate br = baud_19200;

	if (verb)
		printf("Setting up the instrument\n");

	it = new_inst(comport, itype);
	if (debug)
		it->debug = debug;	/* Turn on debugging */
	if (verb)
		it->verb = verb;	/* Turn on verbosity */

	/* Establish communications */
	if ((rv = it->init_coms(it, comport, br, 15.0)) != inst_ok) {
#ifdef DEBUG
		printf("init_coms returned '%s' (%s)\n",
		       it->inst_interp_error(it, rv), it->interp_error(it, rv));
#endif /* DEBUG */
		it->del(it);
		return -1;
	}

	/* Initialise the instrument */
	if ((rv = it->init_inst(it)) != inst_ok) {
#ifdef DEBUG
		printf("init_inst returned '%s' (%s)\n",
		       it->inst_interp_error(it, rv), it->interp_error(it, rv));
#endif /* DEBUG */
		it->del(it);
		return -1;
	}

	/* Set CRT or LCD mode */
	if (dtype == 1 || dtype == 2) {
		inst_opt_mode om;

		if (dtype == 1)
			om = inst_opt_disp_crt;
		else
			om = inst_opt_disp_lcd;

		if ((rv = it->set_opt_mode(it,om)) != inst_ok) {
#ifdef DEBUG
			printf("Setting display type failed failed with '%s' (%s)\n",
			       it->inst_interp_error(it, rv), it->interp_error(it, rv));
#endif /* DEBUG */
			it->del(it);
			return -1;
		}
	}

	/* Disable autocalibration of machine if selected */
	if (nocal != 0){
		if ((rv = it->set_opt_mode(it,inst_opt_noautocalib)) != inst_ok) {
#ifdef DEBUG
			printf("Setting no-autocalibrate failed failed with '%s' (%s)\n",
			       it->inst_interp_error(it, rv), it->interp_error(it, rv));
#endif /* DEBUG */
			it->del(it);
			return -1;
		}
	}

	if (it->itype == instDTP92
	 || it->itype == instDTP94) {
		dispwin *dw;

		/* Offset calibration */
		empty_con_chars();
		printf("Offset calibration:\n");
		printf("Place instrument on a dark surface.\n");
		printf("Hit Esc to skip, any other key to start calibration:"); fflush(stdout);
		if ((c = next_con_char()) == 0x1b) {
			printf("\n");
		} else {
			printf("\n");

			if ((rv = it->calibrate(it, 0)) != inst_ok) {
#ifdef DEBUG
				printf("calibrate returned '%s' (%s)\n",
				       it->inst_interp_error(it, rv), it->interp_error(it, rv));
#endif /* DEBUG */
				printf("Calibrate failed with '%s' (%s)\n",
			       it->inst_interp_error(it, rv), it->interp_error(it, rv));
				it->del(it);
				return -1;
			} else {
				printf("Offset calibration successful\n");
			}

			/* The user may remove the instrument */
			printf("The instrument can be removed from the dark surface.\n");
		}

		if (it->itype == instDTP92) {
			/* Cell ratio calibration */
			/* Open display window */
			if ((dw = new_dispwin(disp, patsize, patsize, ho, vo, 0, 0, override)) == NULL) {
#ifdef DEBUG
				printf("Failed to creat a display window \n");
#endif /* DEBUG */
				it->del(it);
				return -1;
			}
	
			empty_con_chars();
			printf("Cell ratio calibration:\n");
			printf("Place instrument on test window.\n");
			printf("Hit Esc to skip, any other key to start calibration:"); fflush(stdout);
			if ((c = next_con_char()) == 0x1b) {
				printf("\n");
			} else {
				int i;
				double gy;
	
				printf("\n");
	
				for (i = 0, gy = 0.6; i < 3; i++) {
	
					dw->set_color(dw, gy, gy, gy);
	
					if ((rv = it->calibrate(it, 1)) != inst_ok) {
	#ifdef DEBUG
						printf("calibrate returned '%s' (%s)\n",
						       it->inst_interp_error(it, rv), it->interp_error(it, rv));
	#endif /* DEBUG */
						if ((rv & 0xff) == DTP92_TOO_MUCH_LIGHT)
							gy *= 0.7;
						else if ((rv & 0xff) == DTP92_NOT_ENOUGH_LIGHT) {
							gy *= 1.4;
							if (gy > 1.0)
								gy = 1.0;
						} else  {
							printf("Calibrate failed with '%s' (%s)\n",
						       it->inst_interp_error(it, rv), it->interp_error(it, rv));
							dw->del(dw);
							it->del(it);
							return -1;
						}
					} else {
						printf("Cell ratio calibration successful\n");
						break;		/* Succeeded */
					}
	
					/* Try again */
				}
	
				if (i >= 3) {
					printf("Calibration failed - too many tries.\n");
					dw->del(dw);
					it->del(it);
					return -1;
				}
	
				/* Remove sample window */
				dw->del(dw);
	
				/* The user may remove the instrument */
				printf("The instrument can be removed from the screen.\n");
	
			}
		}

	} else if (it->itype == instSpectrolino) {

		if ((rv = it->set_mode(it, inst_mode_emis_disp)) != inst_ok) {
#ifdef DEBUG
			printf("set_mode returned '%s' (%s)\n",
				   it->inst_interp_error(it, rv), it->interp_error(it, rv));
#endif /* DEBUG */
			printf("Set_mode failed with '%s' (%s)\n",
			       it->inst_interp_error(it, rv), it->interp_error(it, rv));
			return -1;
		}

		if ((rv = it->calibrate(it, 0)) != inst_ok) {
			printf("Calibrate failed with '%s' (%s)\n",
		       it->inst_interp_error(it, rv), it->interp_error(it, rv));
			it->del(it);
			return -1;
		}
	}

	/* clean up */
	it->del(it);

	return 0;
}

/* Take a series of readings from the display */
/* Return nz on fail/abort */
/* 1 = user aborted */
/* 2 = instrument access failed */
/* 3 = window access failed */ 
/* 4 = user hit terminate key */
/* 5 = system error */
/* Use disprd_err() to interpret it */
static int disprd_read(disprd *p,
	col *cols,		/* Array of patch colors to be tested */
	int npat, 		/* Number of patches to be tested */
	int spat,		/* Start patch index for "verb", 0 if not used */
	int tpat,		/* Total patch index for "verb", 0 if not used */
	int tc			/* If nz, termination key */
) {
	int j, rv;
	int patch;
	ipatch val;		/* Return value */
	int c;			/* Character */

	/* Setup user termination character */
	p->it->icom->set_user_term(p->it->icom, tc);

	for (patch = 0; patch < npat; patch++) {
		col *scb = &cols[patch];

		scb->XYZ_v = 0;		/* No readings are valid */
		scb->aXYZ_v = 0;
		scb->sp.spec_n = 0;

		if (p->verb && spat != 0 && tpat != 0) {
			fprintf(p->df,"\rpatch %d of %d",spat+patch,tpat);
			fflush(p->df);
		}
#ifdef DEBUG
		printf("About to read patch %d\n",patch);
#endif /* DEBUG */

		if ((rv = p->dw->set_color(p->dw, scb->r, scb->g, scb->b)) != 0) {
#ifdef DEBUG
			printf("set_color() returned %s\n",rv);
#endif /* DEBUG */
			return 3;
		}

		for (;;) {	/* Until we give up retrying */
			val.XYZ_v = 0;		/* No readings are valid */
			val.aXYZ_v = 0;
			val.sp.spec_n = 0;
	
			if ((rv = p->it->read_sample(p->it, scb->id, &val)) != inst_ok) {
#ifdef DEBUG
				printf("read_sample returned '%s' (%s)\n",
			       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
				/* Deal with a user terminate */
				if ((rv & inst_mask) == inst_user_term) {
					return 4;
				/* Deal with a user abort */
				} else if ((rv & inst_mask) == inst_user_abort) {
					empty_con_chars();
					printf("\nSample read stopped at user request!\n");
					printf("Hit Esc to give up, any other key to retry:"); fflush(stdout);
					if ((c = next_con_char()) == 0x1b || c == 0x03) {
						printf("\n");
						return 1;
					}
					printf("\n");
					continue;
				/* Deal with a misread or needs calibration */
				} else if ((rv & inst_mask) == inst_misread || (rv & inst_mask) == inst_needs_cal) {
					empty_con_chars();
					if ((rv & inst_mask) == inst_needs_cal) {
						printf("\nSample read failed because instruments needs calibration\n");
						printf("Hit Esc to give up, or do calibration and then hit any\n");
						printf("other key to retry:"); fflush(stdout);
					} else {
						printf("\nSample read failed due to misread\n");
						printf("Hit Esc to give up, any other key to retry:"); fflush(stdout);
					}
					if ((c = next_con_char()) == 0x1b || c == 0x03) {
						printf("\n");
						return 1;
					}
					printf("\n");
					continue;
				/* Deal with a communications error */
				} else if ((rv & inst_mask) == inst_coms_fail) {
					int tt = p->it->last_sioerr(p->it);
					if (tt & (ICOM_BRK | ICOM_FER | ICOM_PER | ICOM_OER)) {
						if (p->br == baud_19200) p->br = baud_9600;
						else if (p->br == baud_9600) p->br = baud_4800;
						else if (p->br == baud_2400) p->br = baud_1200;
						else p->br = baud_1200;
					}
					/* Communication problem, allow retrying at a lower baud rate */
					empty_con_chars();
					printf("\nSample read failed due to communication problem.\n");
					printf("Hit Esc to give up, any other key to retry:"); fflush(stdout);
					if ((c = next_con_char()) == 0x1b || c == 0x03) {
						printf("\n");
						return 1;
					}
					printf("\n");
					if ((rv = p->it->init_coms(p->it, p->comport, p->br, 15.0)) != inst_ok) {
#ifdef DEBUG
						printf("init_coms returned '%s' (%s)\n",
					       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
						return 2;
					}
				continue;
				}
			} else {
				break;		/* Sucesful reading */
			}
		}
		/* We only fall through with a valid reading */
#ifdef DEBUG
		printf("got reading, transfering to col\n");
#endif /* DEBUG */

		/* Copy relative XYZ */
		if (val.XYZ_v != 0) {
			for (j = 0; j < 3; j++)
				scb->XYZ[j] = val.XYZ[j];
			scb->XYZ_v = 1;
		}

		/* Copy absolute XYZ */
		if (val.aXYZ_v != 0) {
			for (j = 0; j < 3; j++)
				scb->aXYZ[j] = val.aXYZ[j];
			scb->aXYZ_v = 1;
		}

		/* Copy spectral */
		if (p->spectral && val.sp.spec_n > 0) {
			scb->sp = val.sp;
		}
#ifdef DEBUG
		printf("on to next reading\n");
#endif /* DEBUG */
	}
	if (p->verb && spat != 0 && tpat != 0 && (spat+patch-1) == tpat)
		fprintf(p->df,"\n");
	return 0;
}


/* Test without a spectrometer using a fake device */
static int disprd_fake_read(disprd *p,
	col *cols,		/* Array of patch colors to be tested */
	int npat, 		/* Number of patches to be tested */
	int spat,		/* Start patch index for "verb", 0 if not used */
	int tpat,		/* Total patch index for "verb", 0 if not used */
	int tc			/* If nz, termination key */
) {
	icmXYZNumber white;		/* White point */
	icmXYZNumber red;		/* Red colorant */
	icmXYZNumber green;		/* Green colorant */
	icmXYZNumber blue;		/* Blue colorant */
	double doff[3];			/* device offsets */
	double mat[3][3];		/* Destination matrix */
	double ooff[3];			/* XYZ offsets */
	double rgb[3];
	double br = 35.4;		/* Overall brightness */
	int patch, j;
	int ttpat = tpat;

	if (ttpat < npat)
		ttpat = npat;

	/* Setup fake device */
	white.X = br * 0.955;		/* Somewhere between D50 and D65 */
	white.Y = br * 1.00;
	white.Z = br * 0.97;
	red.X = br * 0.41;
	red.Y = br * 0.21;
	red.Z = br * 0.02;
	green.X = br * 0.30;
	green.Y = br * 0.55;
	green.Z = br * 0.15;
	blue.X = br * 0.15;
	blue.Y = br * 0.10;
	blue.Z = br * 0.97;
	/* Input offset, equivalent to RGB offsets having various values */
	doff[0] = 0.10;
	doff[1] = 0.06;
	doff[2] = 0.08;
	/* Output offset - equivalent to flare */
	ooff[0] = 0.04;
	ooff[1] = 0.09;
	ooff[2] = 0.16;

	if (icmRGBprim2matrix(white, red, green, blue, mat))
		error("Fake read unexpectedly got singular matrix\n");

	for (patch = 0; patch < npat; patch++) {
		if (p->verb && spat != 0 && tpat != 0) {
			fprintf(p->df,"\rpatch %d of %d",spat+patch,tpat);
			fflush(p->df);
		}
		rgb[0] = cols[patch].r;
		rgb[1] = cols[patch].g;
		rgb[2] = cols[patch].b;

		/* Apply device offset */
		for (j = 0; j < 3; j++) {
			if (rgb[j] < 0.0)
				rgb[j] = 0.0;
			else if (rgb[j] > 1.0)
				rgb[j] = 1.0;
			rgb[j] = doff[j] + (1.0 - doff[j]) * rgb[j];
		}

		/* Apply gamma */
		for (j = 0; j < 3; j++) {
			if (rgb[j] >= 0.0)
				rgb[j] = pow(rgb[j], 2.5);
			else
				rgb[j] = 0.0;
		}

		/* Convert to XYZ */
		icmMulBy3x3(cols[patch].aXYZ, mat, rgb);

		/* Apply XYZ offset */
		for (j = 0; j < 3; j++)
			cols[patch].aXYZ[j] += ooff[j];

		cols[patch].aXYZ_v = 1;
	}
	if (p->verb && spat != 0 && tpat != 0 && (spat+patch-1) == tpat)
		fprintf(p->df,"\n");
	return 0;
}

/* Return a string describing the error code */
char *disprd_err(int en) {
	switch(en) {
		case 1:
			return "User Aborted";
		case 2:
			return "Instrument Access Failed";
		case 3:
			return "Window Access Failed";
		case 4:
			return "User Terminated";
		case 5:
			return "System Error";
	}
	return "Unknown";
}

static void disprd_del(disprd *p) {

	/* The user may remove the instrument */
	printf("The instrument can be removed from the screen.\n");

	if (p->it != NULL)
		p->it->del(p->it);
	if (p->dw != NULL) {
		if (p->or != NULL)
			p->dw->set_ramdac(p->dw,p->or);
		p->dw->del(p->dw);
	}
	free(p);
}

int new_disprd_errc = 0;

/* Create a display reading object. */
/* Return NULL if error */
/* Set global new_disprd_errc to error code: */
/* 1 = user aborted */
/* 2 = instrument access failed */
/* 3 = window access failed */ 
/* 4 = user hit terminate key */
/* 5 = system error */
/* Use disprd_err() to interpret it */
disprd *new_disprd(
instType itype,		/* Instrument type */
int comport, 		/* COM port used. -99 == fake Display */
int dtype,			/* Display type, 1 = CRT, 2 = LCD */
int nocal,			/* No automatic instrument calibration */
int donat,			/* Use ramdac for native output, else run through current or set ramdac */
double cal[3][256],	/* Calibration set/return (cal[0][0] < 0.0 if can't/not to be used) */
disppath *disp,		/* Display to calibrate. */
int override,		/* Override_redirect on X11 */
double patsize,		/* Size of dispwin */
double ho,			/* Horizontal offset */
double vo,			/* Vertical offset */
int spectral,		/* Generate spectral info flag */
int verb,			/* Verbosity flag */
FILE *df,			/* Verbose output - NULL = stdout */
int debug			/* Debug flag */
) {
	disprd *p;
	int c;
	inst_code rv;
	
	new_disprd_errc = 0;

	/* Allocate a disprd */
	if ((p = (disprd *)calloc(sizeof(disprd), 1)) == NULL) {
		new_disprd_errc = 5;
		return NULL;
	}
	p->del = disprd_del;
	p->read = disprd_read;

	p->verb = verb;
	if (df)
		p->df = df;
	else
		p->df = stdout;
	p->comport = comport;
	p->br = baud_19200;

	if (comport == -99) {
		p->fake = 1;
		p->read = disprd_fake_read;
		return p;
	}

	if (verb)
		fprintf(p->df,"Setting up the instrument\n");

	if ((p->it = new_inst(comport, itype)) == NULL) {
		p->del(p);
		new_disprd_errc = 2;
		return NULL;
	}

	if (debug)
		p->it->debug = debug;	/* Turn on debugging */

	/* Establish communications */
	if ((rv = p->it->init_coms(p->it, p->comport, p->br, 15.0)) != inst_ok) {
#ifdef DEBUG
		printf("init_coms returned '%s' (%s)\n",
		       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
		p->del(p);
		new_disprd_errc = 2;
		return NULL;
	}

	/* Initialise the instrument */
	if ((rv = p->it->init_inst(p->it)) != inst_ok) {
#ifdef DEBUG
		printf("init_inst returned '%s' (%s)\n",
		       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
		p->del(p);
		new_disprd_errc = 2;
		return NULL;
	}

	/* Configure the instrument mode */
	{
		inst_capability cap;
		inst_mode mode = 0;

		cap = p->it->capabilities(p->it);

		if ((cap & inst_emis_disp) == 0) {
			printf("Need emissive measurement capability,\n");
			printf("but instrument doesn't support it\n");
			p->del(p);
			new_disprd_errc = 2;
			return NULL;
		}

		if (spectral && (cap & inst_spectral) == 0) {
			printf("Spectral information was requested,\n");
			printf("but instrument doesn't support it\n");
			spectral = 0;
		}

		mode = inst_mode_emis_disp;

		if (spectral) {
			mode |= inst_mode_spectral;
			p->spectral = 1;
		} else {
			p->spectral = 0;
		}
	

		/* Set CRT or LCD mode */
		if (dtype == 1 || dtype == 2) {
			inst_opt_mode om;

			if (dtype == 1)
				om = inst_opt_disp_crt;
			else
				om = inst_opt_disp_lcd;

			if ((rv = p->it->set_opt_mode(p->it, om)) != inst_ok) {
#ifdef DEBUG
				printf("Setting display type failed failed with '%s' (%s)\n",
			       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
				p->del(p);
				new_disprd_errc = 2;
				return NULL;
			}
		}

		/* Disable autocalibration of machine if selected */
		if (nocal != 0){
			if ((rv = p->it->set_opt_mode(p->it,inst_opt_noautocalib)) != inst_ok) {
#ifdef DEBUG
				printf("Setting no-autocalibrate failed failed with '%s' (%s)\n",
			       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
				p->del(p);
				new_disprd_errc = 2;
				return NULL;
			}
		}

		if ((rv = p->it->set_mode(p->it,mode)) != inst_ok) {
#ifdef DEBUG
			printf("set_mode returned '%s' (%s)\n",
			       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
#endif /* DEBUG */
			p->del(p);
			new_disprd_errc = 2;
			return NULL;
		}

		/* Do a calibration up front, so as not to get in the users way. */
		if (p->it->needs_calibration(p->it) == inst_needs_cal) {
			if ((rv = p->it->calibrate(p->it, 0)) != inst_ok) {
				printf("Calibrate failed with '%s' (%s)\n",
			       p->it->inst_interp_error(p->it, rv), p->it->interp_error(p->it, rv));
				p->del(p);
				new_disprd_errc = 2;
				return NULL;
			}
		}
	}

	/* Open display window */
	if ((p->dw = new_dispwin(disp, patsize, patsize, ho, vo, 0, donat, override)) == NULL) {
#ifdef DEBUG
		printf("Failed to creat a display window \n");
#endif /* DEBUG */
		p->del(p);
		new_disprd_errc = 3;
		return NULL;
	}

	/* Save current RAMDAC so that we can restore it */
	if ((p->or = p->dw->get_ramdac(p->dw)) == NULL) {
		if (p->verb)
			fprintf(p->df,"Unable to read or set display RAMDAC\n");
	}

	/* Set the given RAMDAC so we can characterise through it */
	if (cal != NULL && cal[0][0] >= 0.0 && p->or != NULL) {
		ramdac *r;
		
		r = p->or->clone(p->or);
		if (r->nent != 256) {
			if (p->verb)
				fprintf(p->df,"Can't currently use a RAMDAC which doesn't have 256 entries");
		} else {
			int j, i;
			/* Set the ramdac contents */
			for (j = 0; j < 3; j++) {
				for (i = 0; i < r->nent; i++) {
					r->v[j][i] = cal[j][i];
				}
			}
			if (p->dw->set_ramdac(p->dw, r)) {
				if (p->verb) {
					fprintf(p->df,"Failed to set RAMDAC to desired calibration.\n");
					fprintf(p->df,"Perhaps the operating system is being fussy ?\n");
				}
				r->del(r);
				p->del(p);
				new_disprd_errc = 3;
				return NULL;
			}
			r->del(r);
		}
	}

	/* Return the ramdac being used */
	if (p->or != NULL && cal != NULL) {
		ramdac *r;
		
		if ((r = p->dw->get_ramdac(p->dw)) == NULL) {
			if (p->verb)
				fprintf(p->df,"Failed to read current RAMDAC");
		} else {
			if (r->nent != 256) {
				if (p->verb)
					fprintf(p->df,"Can't currently read a RAMDAC which doesn't have 256 entries");

			} else {
				int j, i;

				/* Get the ramdac contents */
				for (j = 0; j < 3; j++) {
					for (i = 0; i < r->nent; i++) {
						cal[j][i] = r->v[j][i];
					}
				}
			}
			r->del(r);
		}
	}
	
	/* Ask user to put instrument on screen */
	empty_con_chars();
	printf("Place instrument on test window.\n");
	printf("Hit Esc to give up, any other key to continue:"); fflush(stdout);
	if ((c = next_con_char()) == 0x1b || c == 0x03) {
		printf("\n");
		p->del(p);
		new_disprd_errc = 1;
		return NULL;
	}
	printf("\n");

	return p;
}

